/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file contains the necessary initialization of the ILI9341 display
*   controller and the necessary function to update a rectangular area of the
*   screen from an Embedded Wizard UI application.
*   The data transfer is done by using the hardware SPI controller via DMA.
*   The color format of the display is RGB565.
*
*   This driver assumes the following GPIO layout of the interface:
*   Control lines (SPI):
*     MISO - 25
*     MOSI - 23
*     CLK  - 19
*     /CS  - 22               => chip select, low active
*   Additional control lines:
*     DC   - 21               => register select (data/command)
*     /RST - 18               => reset of LCD module, low active
*     /BCKL - 5               => backlight, low active
*
*   This module is just a template - adapt pin configuration and timing (!)
*   according your hardware layout and hardware constraints.
*   Adjust device parameters, initialization and gamma correction values
*   according your display and your needs.
*   This driver was tested on a ESP32-WOVER-KIT display module (3,2" TFT 320x240).
*
*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "esp_heap_caps.h"
#include "esp32s2/rom/lldesc.h"
#include "soc/system_reg.h"
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/spi_master.h"
#include "esp32s2/rom/gpio.h"
#include "esp32s2/rom/ets_sys.h"
#include "esp32s2/rom/ets_sys.h"
#include "soc/periph_defs.h"
#include "soc/spi_struct.h"
#include "hal/gpio_ll.h"

#include "ili9341_spi.h"

/* SPI/LCD GPIO pin definitions */
#define PIN_NUM_MOSI                    9
#define PIN_NUM_CLK                     15
#define PIN_NUM_CS                      11

#define PIN_NUM_DC                      13
#define PIN_NUM_RST                     16
#define PIN_NUM_BCKL                     6

/* display specific settings */
#define DISPLAY_PIXEL_WIDTH             ((uint16_t)320)
#define DISPLAY_PIXEL_HEIGHT            ((uint16_t)240)

/* ILI9341 commands */
#define ILI9341_SWRESET                 0x01   /* Software Reset */
#define ILI9341_READ_DISPLAY_ID         0x04   /* Read display identification information */
#define ILI9341_RDDST                   0x09   /* Read Display Status */
#define ILI9341_RDDPM                   0x0A   /* Read Display Power Mode */
#define ILI9341_RDDMADCTL               0x0B   /* Read Display MADCTL */
#define ILI9341_RDDCOLMOD               0x0C   /* Read Display Pixel Format */
#define ILI9341_RDDIM                   0x0D   /* Read Display Image Format */
#define ILI9341_RDDSM                   0x0E   /* Read Display Signal Mode */
#define ILI9341_RDDSDR                  0x0F   /* Read Display Self-Diagnostic Result */
#define ILI9341_SPLIN                   0x10   /* Enter Sleep Mode */
#define ILI9341_SLEEP_OUT               0x11   /* Sleep out register */
#define ILI9341_PTLON                   0x12   /* Partial Mode ON */
#define ILI9341_NORMAL_MODE_ON          0x13   /* Normal Display Mode ON */
#define ILI9341_DINVOFF                 0x20   /* Display Inversion OFF */
#define ILI9341_DINVON                  0x21   /* Display Inversion ON */
#define ILI9341_GAMMA                   0x26   /* Gamma register */
#define ILI9341_DISPLAY_OFF             0x28   /* Display off register */
#define ILI9341_DISPLAY_ON              0x29   /* Display on register */
#define ILI9341_COLUMN_ADDR             0x2A   /* Colomn address register */
#define ILI9341_PAGE_ADDR               0x2B   /* Page address register */
#define ILI9341_GRAM                    0x2C   /* GRAM register */
#define ILI9341_RGBSET                  0x2D   /* Color SET */
#define ILI9341_RAMRD                   0x2E   /* Memory Read */
#define ILI9341_PLTAR                   0x30   /* Partial Area */
#define ILI9341_VSCRDEF                 0x33   /* Vertical Scrolling Definition */
#define ILI9341_TEOFF                   0x34   /* Tearing Effect Line OFF */
#define ILI9341_TEON                    0x35   /* Tearing Effect Line ON */
#define ILI9341_MAC                     0x36   /* Memory Access Control register*/
#define ILI9341_VSCRSADD                0x37   /* Vertical Scrolling Start Address */
#define ILI9341_IDMOFF                  0x38   /* Idle Mode OFF */
#define ILI9341_IDMON                   0x39   /* Idle Mode ON */
#define ILI9341_PIXEL_FORMAT            0x3A   /* Pixel Format register */
#define ILI9341_WRITE_MEM_CONTINUE      0x3C   /* Write Memory Continue */
#define ILI9341_READ_MEM_CONTINUE       0x3E   /* Read Memory Continue */
#define ILI9341_SET_TEAR_SCANLINE       0x44   /* Set Tear Scanline */
#define ILI9341_GET_SCANLINE            0x45   /* Get Scanline */
#define ILI9341_WDB                     0x51   /* Write Brightness Display register */
#define ILI9341_RDDISBV                 0x52   /* Read Display Brightness */
#define ILI9341_WCD                     0x53   /* Write Control Display register*/
#define ILI9341_RDCTRLD                 0x54   /* Read CTRL Display */
#define ILI9341_WRCABC                  0x55   /* Write Content Adaptive Brightness Control */
#define ILI9341_RDCABC                  0x56   /* Read Content Adaptive Brightness Control */
#define ILI9341_WRITE_CABC              0x5E   /* Write CABC Minimum Brightness */
#define ILI9341_READ_CABC               0x5F   /* Read CABC Minimum Brightness */
#define ILI9341_READ_ID1                0xDA   /* Read ID1 */
#define ILI9341_READ_ID2                0xDB   /* Read ID2 */
#define ILI9341_READ_ID3                0xDC   /* Read ID3 */
#define ILI9341_RGB_INTERFACE           0xB0   /* RGB Interface Signal Control */
#define ILI9341_FRMCTR1                 0xB1   /* Frame Rate Control (In Normal Mode) */
#define ILI9341_FRMCTR2                 0xB2   /* Frame Rate Control (In Idle Mode) */
#define ILI9341_FRMCTR3                 0xB3   /* Frame Rate Control (In Partial Mode) */
#define ILI9341_INVTR                   0xB4   /* Display Inversion Control */
#define ILI9341_BPC                     0xB5   /* Blanking Porch Control register */
#define ILI9341_DFC                     0xB6   /* Display Function Control register */
#define ILI9341_ETMOD                   0xB7   /* Entry Mode Set */
#define ILI9341_BACKLIGHT1              0xB8   /* Backlight Control 1 */
#define ILI9341_BACKLIGHT2              0xB9   /* Backlight Control 2 */
#define ILI9341_BACKLIGHT3              0xBA   /* Backlight Control 3 */
#define ILI9341_BACKLIGHT4              0xBB   /* Backlight Control 4 */
#define ILI9341_BACKLIGHT5              0xBC   /* Backlight Control 5 */
#define ILI9341_BACKLIGHT7              0xBE   /* Backlight Control 7 */
#define ILI9341_BACKLIGHT8              0xBF   /* Backlight Control 8 */
#define ILI9341_POWER1                  0xC0   /* Power Control 1 register */
#define ILI9341_POWER2                  0xC1   /* Power Control 2 register */
#define ILI9341_VCOM1                   0xC5   /* VCOM Control 1 register */
#define ILI9341_VCOM2                   0xC7   /* VCOM Control 2 register */
#define ILI9341_NVMWR                   0xD0   /* NV Memory Write */
#define ILI9341_NVMPKEY                 0xD1   /* NV Memory Protection Key */
#define ILI9341_RDNVM                   0xD2   /* NV Memory Status Read */
#define ILI9341_READ_ID4                0xD3   /* Read ID4 */
#define ILI9341_PGAMMA                  0xE0   /* Positive Gamma Correction register */
#define ILI9341_NGAMMA                  0xE1   /* Negative Gamma Correction register */
#define ILI9341_DGAMCTRL1               0xE2   /* Digital Gamma Control 1 */
#define ILI9341_DGAMCTRL2               0xE3   /* Digital Gamma Control 2 */
#define ILI9341_INTERFACE               0xF6   /* Interface control register */
#define ILI9341_POWERA                  0xCB   /* Power control A register */
#define ILI9341_POWERB                  0xCF   /* Power control B register */
#define ILI9341_DTCA                    0xE8   /* Driver timing control A */
#define ILI9341_DTCB                    0xEA   /* Driver timing control B */
#define ILI9341_POWER_SEQ               0xED   /* Power on sequence register */
#define ILI9341_3GAMMA_EN               0xF2   /* 3 Gamma enable register */
#define ILI9341_PRC                     0xF7   /* Pump ratio control register */

#define LCD_DMA_MAX_SIZE                (4095)

typedef struct {
  uint32_t clk_fre;
  uint8_t pin_clk;
  uint8_t pin_mosi;
  uint8_t pin_dc;
  uint8_t pin_cs;
  uint8_t pin_rst;
  uint8_t pin_bk;
  uint8_t horizontal;
  uint32_t max_buffer_size; // DMA used
  bool swap_data;
} lcd_config_t;

typedef struct {
  uint32_t buffer_size;
  uint32_t half_buffer_size;
  uint32_t node_cnt;
  uint32_t half_node_cnt;
  uint32_t dma_size;
  uint8_t pin_dc;
  uint8_t pin_cs;
  uint8_t pin_rst;
  uint8_t pin_bk;
  lldesc_t *dma;
  uint8_t *buffer;
  bool swap_data;
  QueueHandle_t event_queue;
} lcd_obj_t;

static lcd_obj_t *lcd_obj = NULL;

/* helper function to set the D/C line according the flag provided within user field */
/* Attention: function is called in context of SPI interrupt */
static void IRAM_ATTR lcd_dma_isr(void *arg)
{
  BaseType_t HPTaskAwoken = pdFALSE;
  typeof(GPSPI3.dma_int_st) int_st = GPSPI3.dma_int_st;
  GPSPI3.dma_int_clr.val = int_st.val;

  if (int_st.out_eof) {
    xQueueSendFromISR(lcd_obj->event_queue, (void *)&int_st.val, &HPTaskAwoken);
  }

  if (HPTaskAwoken == pdTRUE) {
    portYIELD_FROM_ISR();
  }
}

static void spi_write_data(uint8_t *data, size_t len)
{
  int event  = 0;
  int x = 0, cnt = 0, size = 0;
  int end_pos = 0;

  /*!< Generate a data DMA linked list */
  for (x = 0; x < lcd_obj->node_cnt; x++) {
    lcd_obj->dma[x].size = lcd_obj->dma_size;
    lcd_obj->dma[x].length = lcd_obj->dma_size;
    lcd_obj->dma[x].buf = (lcd_obj->buffer + lcd_obj->dma_size * x);
    lcd_obj->dma[x].eof = !((x + 1) % lcd_obj->half_node_cnt);
    lcd_obj->dma[x].empty = (uint32_t)&lcd_obj->dma[(x + 1) % lcd_obj->node_cnt];
  }

  lcd_obj->dma[lcd_obj->half_node_cnt - 1].empty = 0;
  lcd_obj->dma[lcd_obj->node_cnt - 1].empty = 0;
  cnt = len / lcd_obj->half_buffer_size;
  /*!< Start the signal */
  xQueueSend(lcd_obj->event_queue, &event, 0);

  /*!< Processing a complete piece of data, ping-pong operation */
  for (x = 0; x < cnt; x++) {
    uint8_t *out = (uint8_t *)lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt].buf;
    if (lcd_obj->swap_data) {
      for (size_t i = 0; i < lcd_obj->half_buffer_size; i += 2) {
        out[i]   = data[i+1];
        out[i+1] = data[i];
      }
    } else {
      memcpy(out, data, lcd_obj->half_buffer_size);
    }

    data += lcd_obj->half_buffer_size;
    xQueueReceive(lcd_obj->event_queue, (void *)&event, portMAX_DELAY);
    GPSPI3.mosi_dlen.usr_mosi_bit_len = lcd_obj->half_buffer_size * 8 - 1;
    GPSPI3.dma_out_link.addr = ((uint32_t)&lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt]) & 0xfffff;
    GPSPI3.dma_out_link.start = 1;
    ets_delay_us(1);
    GPSPI3.cmd.usr = 1;
  }

  cnt = len % lcd_obj->half_buffer_size;

  /*!< Processing remaining incomplete segment data */
  if (cnt) {
    uint8_t *out = (uint8_t *)lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt].buf;
    if (lcd_obj->swap_data && cnt > 1) {
      for (size_t i = 0; i < cnt; i += 2) {
        out[i]   = data[i+1];
        out[i+1] = data[i];
      }
    } else {
      memcpy(out, data, cnt);
    }

    /*!< Handle the case where the data length is an integer multiple of lcd_obj->dma_size */
    if (cnt % lcd_obj->dma_size) {
      end_pos = (x % 2) * lcd_obj->half_node_cnt + cnt / lcd_obj->dma_size;
      size = cnt % lcd_obj->dma_size;
    } else {
      end_pos = (x % 2) * lcd_obj->half_node_cnt + cnt / lcd_obj->dma_size - 1;
      size = lcd_obj->dma_size;
    }

    /*!< Handle the tail node to make it a DMA tail */
    lcd_obj->dma[end_pos].size = size;
    lcd_obj->dma[end_pos].length = size;
    lcd_obj->dma[end_pos].eof = 1;
    lcd_obj->dma[end_pos].empty = 0;
    xQueueReceive(lcd_obj->event_queue, (void *)&event, portMAX_DELAY);
    GPSPI3.mosi_dlen.usr_mosi_bit_len = cnt * 8 - 1;
    GPSPI3.dma_out_link.addr = ((uint32_t)&lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt]) & 0xfffff;
    GPSPI3.dma_out_link.start = 1;
    ets_delay_us(1);
    GPSPI3.cmd.usr = 1;
  }

  xQueueReceive(lcd_obj->event_queue, (void *)&event, portMAX_DELAY);
}


/* helper function to write a command register value in polling mode */
static void ILI9341_WriteCmnd( uint8_t aCmnd )
{
  gpio_set_level(lcd_obj->pin_dc, 0);
  spi_write_data(&aCmnd, 1);
}


/* helper function to read a data byte in polling mode */
static uint8_t ILI9341_ReadByte( void )
{
  return 0;
}


/* helper function to write a data byte in polling mode */
static void ILI9341_WriteByte( uint8_t aData )
{
  gpio_set_level(lcd_obj->pin_dc, 1);
  spi_write_data(&aData, 1);
}


/* helper function to write a data byte in polling mode */
static void ILI9341_WriteBytes(uint8_t *data, size_t len)
{
  gpio_set_level(lcd_obj->pin_dc, 1);
  spi_write_data(data, len);
}


/* helper function to initialize reset and backlight pins of LCD and to perform
   proper reset of display */
static void ILI9341_ResetDisplay()
{
  /* disable backlight */
  gpio_set_level( PIN_NUM_BCKL, 1 );

  /* apply hardware reset */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 5 / portTICK_PERIOD_MS );   /* reset signal asserted during 5ms  */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 20 / portTICK_PERIOD_MS );  /* reset signal asserted during 20ms */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_Init
*
* DESCRIPTION:
*   The function initializes the SPI hardware and all necessary GPIO in order
*   to reset and intialize the connected display hardware.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
static void lcd_config(lcd_config_t *config)
{
  REG_CLR_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_RST);
  REG_CLR_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_RST);
  REG_CLR_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_DMA_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_DMA_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_DMA_RST);
  REG_CLR_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_DMA_RST);

  int div = 2;

  if (config->clk_fre == 80000000) {
    GPSPI3.clock.clk_equ_sysclk = 1;
  } else {
    GPSPI3.clock.clk_equ_sysclk = 0;
    div = 80000000 / config->clk_fre;
  }

  GPSPI3.ctrl1.clk_mode = 0;
  GPSPI3.clock.clkdiv_pre = 1 - 1;
  GPSPI3.clock.clkcnt_n = div - 1;
  GPSPI3.clock.clkcnt_l = div - 1;
  GPSPI3.clock.clkcnt_h = ((div >> 1) - 1);

  GPSPI3.misc.ck_dis = 0;

  GPSPI3.user1.val = 0;
  GPSPI3.slave.val = 0;
  GPSPI3.misc.ck_idle_edge = 0;
  GPSPI3.user.ck_out_edge = 0;
  GPSPI3.ctrl.wr_bit_order = 0;
  GPSPI3.ctrl.rd_bit_order = 0;
  GPSPI3.user.val = 0;
  GPSPI3.user.cs_setup = 1;
  GPSPI3.user.cs_hold = 1;
  GPSPI3.user.usr_mosi = 1;
  GPSPI3.user.usr_mosi_highpart = 0;

  GPSPI3.dma_conf.val = 0;
  GPSPI3.dma_conf.out_rst = 1;
  GPSPI3.dma_conf.out_rst = 0;
  GPSPI3.dma_conf.ahbm_fifo_rst = 1;
  GPSPI3.dma_conf.ahbm_fifo_rst = 0;
  GPSPI3.dma_conf.ahbm_rst = 1;
  GPSPI3.dma_conf.ahbm_rst = 0;
  GPSPI3.dma_out_link.dma_tx_ena = 1;
  GPSPI3.dma_conf.out_eof_mode = 1;
  GPSPI3.cmd.usr = 0;

  GPSPI3.dma_int_clr.val = ~0;
  GPSPI3.dma_int_ena.val = 0;
  GPSPI3.dma_int_ena.out_eof = 1;

  intr_handle_t intr_handle = NULL;
  esp_intr_alloc(ETS_SPI3_DMA_INTR_SOURCE, 0, lcd_dma_isr, NULL, &intr_handle);
}

static void lcd_set_pin(lcd_config_t *config)
{
  PIN_FUNC_SELECT(GPIO_PIN_MUX_REG[config->pin_clk], PIN_FUNC_GPIO);
  gpio_set_direction(config->pin_clk, GPIO_MODE_OUTPUT);
  gpio_set_pull_mode(config->pin_clk, GPIO_FLOATING);
  gpio_matrix_out(config->pin_clk, SPI3_CLK_OUT_MUX_IDX, 0, 0);

  PIN_FUNC_SELECT(GPIO_PIN_MUX_REG[config->pin_mosi], PIN_FUNC_GPIO);
  gpio_set_direction(config->pin_mosi, GPIO_MODE_OUTPUT);
  gpio_set_pull_mode(config->pin_mosi, GPIO_FLOATING);
  gpio_matrix_out(config->pin_mosi, SPI3_D_OUT_IDX, 0, 0);

  /*!< Initialize non-SPI GPIOs */
  gpio_config_t io_conf;
  io_conf.intr_type = GPIO_PIN_INTR_DISABLE;
  io_conf.mode = GPIO_MODE_OUTPUT;
  io_conf.pin_bit_mask = (1ULL << config->pin_dc) | (1ULL << config->pin_cs);

  if (config->pin_rst <= 46) {
    io_conf.pin_bit_mask |= (1ULL << config->pin_rst);
  }

  if (config->pin_bk <= 46) {
    io_conf.pin_bit_mask |= (1ULL << config->pin_bk);
  }

  io_conf.pull_down_en = 0;
  io_conf.pull_up_en = 0;
  gpio_config(&io_conf);
}

void lcd_dma_config(lcd_config_t *config)
{
  int cnt = 0;

  if (config->max_buffer_size >= LCD_DMA_MAX_SIZE * 2) {
    lcd_obj->dma_size = LCD_DMA_MAX_SIZE;

    for (cnt = 0;; cnt++) { /*!< Find the buffer size that is divisible by dma_size */
      if ((config->max_buffer_size - cnt) % lcd_obj->dma_size == 0) {
        break;
      }
    }

    lcd_obj->buffer_size = config->max_buffer_size - cnt;
  } else {
    lcd_obj->dma_size = config->max_buffer_size / 2;
    lcd_obj->buffer_size = lcd_obj->dma_size * 2;
  }

  lcd_obj->half_buffer_size = lcd_obj->buffer_size / 2;

  lcd_obj->node_cnt = (lcd_obj->buffer_size) / lcd_obj->dma_size; /*!< Number of DMA nodes */
  lcd_obj->half_node_cnt = lcd_obj->node_cnt / 2;

  lcd_obj->dma  = (lldesc_t *)heap_caps_malloc(lcd_obj->node_cnt * sizeof(lldesc_t), MALLOC_CAP_DMA);
  lcd_obj->buffer = (uint8_t *)heap_caps_malloc(lcd_obj->buffer_size * sizeof(uint8_t), MALLOC_CAP_DMA);
}

int lcd_init(lcd_config_t *config)
{
  lcd_obj = (lcd_obj_t *)heap_caps_calloc(1, sizeof(lcd_obj_t), MALLOC_CAP_DMA);

  lcd_set_pin(config);
  lcd_config(config);
  lcd_dma_config(config);

  lcd_obj->event_queue = xQueueCreate(1, sizeof(int));

  lcd_obj->buffer_size = config->max_buffer_size;
  lcd_obj->swap_data = config->swap_data;

  lcd_obj->pin_dc = config->pin_dc;
  lcd_obj->pin_cs = config->pin_cs;
  lcd_obj->pin_rst = config->pin_rst;
  lcd_obj->pin_bk = config->pin_bk;
  gpio_set_level(lcd_obj->pin_cs, 1);

  /* perform reset of LCD */
  ILI9341_ResetDisplay();

  /* select chip */
  gpio_set_level(lcd_obj->pin_cs, 0);

  /* switch display off */
  ILI9341_WriteCmnd( ILI9341_DISPLAY_OFF );

  /* power control B */
  ILI9341_WriteCmnd( ILI9341_POWERB );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x83 ); // or 0xC1 or 0x81 ???
  ILI9341_WriteByte( 0x30 );

  /* power on sequence control */
  ILI9341_WriteCmnd( ILI9341_POWER_SEQ );
  ILI9341_WriteByte( 0x64 );
  ILI9341_WriteByte( 0x03 );
  ILI9341_WriteByte( 0x12 );
  ILI9341_WriteByte( 0x81 );

  /* driver timing control A */
  ILI9341_WriteCmnd( ILI9341_DTCA );
  ILI9341_WriteByte( 0x85 );
  ILI9341_WriteByte( 0x01 );  // or 0x00
  ILI9341_WriteByte( 0x79 );  // or 0x78

  /* power control A */
  ILI9341_WriteCmnd( ILI9341_POWERA );
  ILI9341_WriteByte( 0x39 );
  ILI9341_WriteByte( 0x2C );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x34 );
  ILI9341_WriteByte( 0x02 );

  /* pump ratio control */
  ILI9341_WriteCmnd( ILI9341_PRC );
  ILI9341_WriteByte( 0x20 );

  /* driver timing control B */
  ILI9341_WriteCmnd( ILI9341_DTCB );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );

  /* power control 1 */
  ILI9341_WriteCmnd( ILI9341_POWER1 );
  ILI9341_WriteByte( 0x26 );    // = 4.75 V

  /* power control 2 */
  ILI9341_WriteCmnd( ILI9341_POWER2 );
  ILI9341_WriteByte( 0x11 );  // or 0x10

  /* VCOM 1 / 2 registers */
  ILI9341_WriteCmnd( ILI9341_VCOM1 );
  ILI9341_WriteByte( 0x35 );  // or 0x45
  ILI9341_WriteByte( 0x3E );  // or 0x15
  ILI9341_WriteCmnd( ILI9341_VCOM2 );
  ILI9341_WriteByte( 0xBE );  // or 0x90

  /* memory access control */
  ILI9341_WriteCmnd( ILI9341_MAC );
  ILI9341_WriteByte( 0x28 ); // or 0x48 or 0xC8 ???

  /* pixel format set */
  ILI9341_WriteCmnd( ILI9341_PIXEL_FORMAT );
  ILI9341_WriteByte( 0x55 );

  /* frame rate */
  ILI9341_WriteCmnd( ILI9341_FRMCTR1 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x1B );

  /* gamma configuration */
  ILI9341_WriteCmnd( ILI9341_3GAMMA_EN );
  ILI9341_WriteByte( 0x08 );

  ILI9341_WriteCmnd( ILI9341_GAMMA );
  ILI9341_WriteByte( 0x01 );

  ILI9341_WriteCmnd( ILI9341_PGAMMA );
  ILI9341_WriteByte( 0x1F );
  ILI9341_WriteByte( 0x1A );
  ILI9341_WriteByte( 0x18 );
  ILI9341_WriteByte( 0x0A );
  ILI9341_WriteByte( 0x0F );
  ILI9341_WriteByte( 0x06 );
  ILI9341_WriteByte( 0x45 );
  ILI9341_WriteByte( 0x87 );
  ILI9341_WriteByte( 0x32 );
  ILI9341_WriteByte( 0x0A );
  ILI9341_WriteByte( 0x07 );
  ILI9341_WriteByte( 0x02 );
  ILI9341_WriteByte( 0x07 );
  ILI9341_WriteByte( 0x05 );
  ILI9341_WriteByte( 0x00 );

  ILI9341_WriteCmnd( ILI9341_NGAMMA );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x25 );
  ILI9341_WriteByte( 0x27 );
  ILI9341_WriteByte( 0x05 );
  ILI9341_WriteByte( 0x10 );
  ILI9341_WriteByte( 0x09 );
  ILI9341_WriteByte( 0x3A );
  ILI9341_WriteByte( 0x78 );
  ILI9341_WriteByte( 0x4D );
  ILI9341_WriteByte( 0x05 );
  ILI9341_WriteByte( 0x18 );
  ILI9341_WriteByte( 0x0D );
  ILI9341_WriteByte( 0x38 );
  ILI9341_WriteByte( 0x3A );
  ILI9341_WriteByte( 0x1F );

  /* colomn address set */
  ILI9341_WriteCmnd( ILI9341_COLUMN_ADDR );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0xEF );

  /* page address set */
  ILI9341_WriteCmnd( ILI9341_PAGE_ADDR );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x01 );
  ILI9341_WriteByte( 0x3F );

  /* entry mode set */
  ILI9341_WriteCmnd( ILI9341_ETMOD );
  ILI9341_WriteByte( 0x07 );

  /* display function control set */
  ILI9341_WriteCmnd( ILI9341_DFC );
  ILI9341_WriteByte( 0x0A );
  ILI9341_WriteByte( 0x82 );
  ILI9341_WriteByte( 0x27 );
  ILI9341_WriteByte( 0x00 );

  /* wake up */
  ILI9341_WriteCmnd( ILI9341_SLEEP_OUT );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* switch display on */
  ILI9341_WriteCmnd( ILI9341_DISPLAY_ON );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* start writing to GRAM */
  ILI9341_WriteCmnd( ILI9341_GRAM );

  /* enable backlight */
  gpio_set_level( PIN_NUM_BCKL, 0 );

  return 0;
}

void ILI9341_Init( void )
{
  lcd_config_t lcd_config = {
    .clk_fre     = 40 * 1000 * 1000,
    .pin_clk     = PIN_NUM_CLK,
    .pin_mosi    = PIN_NUM_MOSI,
    .pin_dc      = PIN_NUM_DC,
    .pin_cs      = PIN_NUM_CS,
    .pin_rst     = PIN_NUM_RST,
    .pin_bk      = PIN_NUM_BCKL,
    .max_buffer_size = 2 * 1024,
    .swap_data     = 0,
  };

  lcd_init(&lcd_config);
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_ReadID
*
* DESCRIPTION:
*   The function assumes a connected and intialized ILI9341 display and tries
*   to read its ID.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   If successful, the function returns the ID of the display - 0 otherwise.
*
*******************************************************************************/
uint16_t ILI9341_ReadID(void)
{
  uint16_t result;

  /* send command */
  ILI9341_WriteCmnd( ILI9341_READ_ID4 );

  /* read dummy data */
  ILI9341_ReadByte();

  /* read register values */
  ILI9341_ReadByte();
  result = ILI9341_ReadByte() << 8;
  result |= ILI9341_ReadByte();

  return result;
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_WaitForCompletion
*
* DESCRIPTION:
*   The function ensures that the previous SPI data transfer is completed..
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ILI9341_WaitForCompletion( void )
{
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_WriteDataWindow
*
* DESCRIPTION:
*   The function sets the destination position and size within the framebuffer
*   of the display according the given rectangle and starts the transfer of
*   the given data. The function does NOT wait for the completion of the transfer.
*
* ARGUMENTS:
*   aX       - Horizontal position for next write access.
*   aY       - Vertical position for next write access.
*   aW       - Width of the data window in pixel.
*   aH       - Height of the data window in pixel.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ILI9341_WriteDataWindow( uint16_t aX, uint16_t aY, uint16_t aW, uint16_t aH, uint8_t* aData )
{
  int       xEnd = aX + aW - 1;
  int       yEnd = aY + aH - 1;

  if ( xEnd >= DISPLAY_PIXEL_WIDTH )
    xEnd = DISPLAY_PIXEL_WIDTH - 1;
  if ( yEnd >= DISPLAY_PIXEL_HEIGHT )
    yEnd = DISPLAY_PIXEL_HEIGHT - 1;

  /* ensure that previous update is completed */
  ILI9341_WaitForCompletion();

  /* colomn address set */
  ILI9341_WriteCmnd(ILI9341_COLUMN_ADDR);
  ILI9341_WriteByte(aX >> 8);
  ILI9341_WriteByte(aX & 0xFF);
  ILI9341_WriteByte(xEnd >> 8);
  ILI9341_WriteByte(xEnd & 0xFF);

  /* page address set */
  ILI9341_WriteCmnd(ILI9341_PAGE_ADDR);
  ILI9341_WriteByte(aY >> 8);
  ILI9341_WriteByte(aY & 0xFF);
  ILI9341_WriteByte(yEnd >> 8);
  ILI9341_WriteByte(yEnd & 0xFF);

  /* start writing to GRAM */
  ILI9341_WriteCmnd(ILI9341_GRAM);
  ILI9341_WriteBytes((uint8_t *)aData, aW * aH * sizeof(uint16_t));
}


/* tzi, msy */