/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file contains the necessary initialization of the ST7789V display
*   controller and the necessary function to update a rectangular area of the
*   screen from an Embedded Wizard UI application.
*   The data transfer is done by using the hardware SPI controller via DMA.
*   The color format of the display is RGB565.
*
*   This driver assumes the following GPIO layout of the interface:
*   Control lines (SPI):
*     MISO - 25
*     MOSI - 23
*     CLK  - 19
*     /CS  - 22               => chip select, low active
*   Additional control lines:
*     DC   - 21               => register select (data/command)
*     /RST - 18               => reset of LCD module, low active
*     /BCKL - 5               => backlight, low active
*
*   This module is just a template - adapt pin configuration and timing (!)
*   according your hardware layout and hardware constraints.
*   Adjust device parameters, initialization and gamma correction values
*   according your display and your needs.
*   This driver was tested on a ESP32-WOVER-KIT display module (3,2" TFT 320x240).
*
*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "esp_heap_caps.h"
#include "esp32s2/rom/lldesc.h"
#include "soc/system_reg.h"
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/spi_master.h"
#include "esp32s2/rom/gpio.h"
#include "esp32s2/rom/ets_sys.h"
#include "esp32s2/rom/ets_sys.h"
#include "soc/periph_defs.h"
#include "soc/spi_struct.h"
#include "hal/gpio_ll.h"

#include "st7789v_spi.h"

/* SPI/LCD GPIO pin definitions */
#define PIN_NUM_MOSI                    9
#define PIN_NUM_CLK                     15
#define PIN_NUM_CS                      11

#define PIN_NUM_DC                      13
#define PIN_NUM_RST                     16
#define PIN_NUM_BCKL                     6

/* display specific settings */
#define DISPLAY_PIXEL_WIDTH             ((uint16_t)320)
#define DISPLAY_PIXEL_HEIGHT            ((uint16_t)240)


/* ST7789V commands */
#define ST7789V_SWRESET                 0x01   /* Software Reset */
#define ST7789V_RDDID                   0x04   /* Read Display ID Information */
#define ST7789V_RDDST                   0x09   /* Read Display Status */
#define ST7789V_RDDPM                   0x0A   /* Read Display Power Mode */
#define ST7789V_RDDMADCTL               0x0B   /* Read Display MADCTL */
#define ST7789V_RDDCOLMOD               0x0C   /* Read Display Pixel Format */
#define ST7789V_RDDIM                   0x0D   /* Read Display Image Format */
#define ST7789V_RDDSM                   0x0E   /* Read Display Signal Mode */
#define ST7789V_RDDSDR                  0x0F   /* Read Display Self-Diagnostic Result */
#define ST7789V_SPLIN                   0x10   /* Enter Sleep Mode */
#define ST7789V_SLPOUT                  0x11   /* Sleep out */
#define ST7789V_PTLON                   0x12   /* Partial Mode ON */
#define ST7789V_NORON                   0x13   /* Normal Display Mode ON */
#define ST7789V_INVOFF                  0x20   /* Display Inversion OFF */
#define ST7789V_INVON                   0x21   /* Display Inversion ON */
#define ST7789V_GAMSET                  0x26   /* Gamma register */
#define ST7789V_DISPOFF                 0x28   /* Display off */
#define ST7789V_DISPON                  0x29   /* Display on */
#define ST7789V_CASET                   0x2A   /* Colomn address */
#define ST7789V_RASET                   0x2B   /* Row address */
#define ST7789V_RAMWR                   0x2C   /* Memory Write */
#define ST7789V_RAMRD                   0x2E   /* Memory Read */
#define ST7789V_PLTAR                   0x30   /* Partial Area */
#define ST7789V_VSCRDEF                 0x33   /* Vertical Scrolling Definition */
#define ST7789V_TEOFF                   0x34   /* Tearing Effect Line OFF */
#define ST7789V_TEON                    0x35   /* Tearing Effect Line ON */
#define ST7789V_MADCTL                  0x36   /* Memory Access Control */
#define ST7789V_VSCSAD                  0x37   /* Vertical Scrolling Start Address */
#define ST7789V_IDMOFF                  0x38   /* Idle Mode OFF */
#define ST7789V_IDMON                   0x39   /* Idle Mode ON */
#define ST7789V_COLMOD                  0x3A   /* Pixel Format register */
#define ST7789V_WRMEMC                  0x3C   /* Write Memory Continue */
#define ST7789V_RDMEMC                  0x3E   /* Read Memory Continue */
#define ST7789V_STE                     0x44   /* Set Tear Scanline */
#define ST7789V_GSCAN                   0x45   /* Get Scanline */
#define ST7789V_WRDISBV                 0x51   /* Write Brightness Display */
#define ST7789V_RDDISBV                 0x52   /* Read Display Brightness */
#define ST7789V_WRCTRLD                 0x53   /* Write Control Display */
#define ST7789V_RDCTRLD                 0x54   /* Read CTRL Display */
#define ST7789V_WRCACE                  0x55   /* Write Content Adaptive Brightness Control */
#define ST7789V_RDCABC                  0x56   /* Read Content Adaptive Brightness Control */
#define ST7789V_WRCABCMB                0x5E   /* Write CABC Minimum Brightness */
#define ST7789V_RDCABCMB                0x5F   /* Read CABC Minimum Brightness */
#define ST7789V_RDABCSDR                0x68   /* Read Automatic Brightness Control Self Diagnostic Result */
#define ST7789V_READ_ID1                0xDA   /* Read ID1 */
#define ST7789V_READ_ID2                0xDB   /* Read ID2 */
#define ST7789V_READ_ID3                0xDC   /* Read ID3 */
#define ST7789V_RAMCTRL                 0xB0   /* RAM Control */
#define ST7789V_RGBCTRL                 0xB1   /* RGB Interface Control */
#define ST7789V_PORCTRL                 0xB2   /* Porch Setting */
#define ST7789V_FRCTRL1                 0xB3   /* Frame Rate Control (In Partial Mode) */
#define ST7789V_PARCTRL                 0xB5   /* Partial Mode Control */
#define ST7789V_GCTRL                   0xB7   /* Gate Control */
#define ST7789V_GTADJ                   0xB8   /* Gate On Timing Adjustment */
#define ST7789V_DGMEN                   0xBA   /* Digital Gamma Enable */
#define ST7789V_VCOMS                   0xBB   /* VCOMS Settings */
#define ST7789V_LCMCTRL                 0xC0   /* LCM Control */
#define ST7789V_IDSET                   0xC1   /* ID Code Setting */
#define ST7789V_VDVVRHEN                0xC2   /* VDV and VRH Command Enable */
#define ST7789V_VRHS                    0xC3   /* VRH Set */
#define ST7789V_VDVS                    0xC4   /* VDV Set */
#define ST7789V_VCMOFSET                0xC5   /* VCOMS Offset Set */
#define ST7789V_FRCTRL2                 0xC6   /* Frame Rate Control in Normal Mode */
#define ST7789V_CABCCTRL                0xC7   /* CABC Control */
#define ST7789V_REGSEL1                 0xC8   /* Register Value Selection 1 */
#define ST7789V_REGSEL2                 0xCA   /* Register Value Selection 2 */
#define ST7789V_PWMFRSEL                0xCC   /* PWM Frequency Selection */
#define ST7789V_PWCTRL1                 0xD0   /* Power Control 1 */
#define ST7789V_VAPVANEN                0xD2   /* Enable VAP/VAN singal output */
#define ST7789V_CMD2EN                  0xDF   /* Command 2 Enable */
#define ST7789V_PVGAMCTRL               0xE0   /* Positive Gamma Correction */
#define ST7789V_NVGAMCTRL               0xE1   /* Negative Gamma Correction */
#define ST7789V_DGMLUTR                 0xE2   /* Digital Gamma LUT Red */
#define ST7789V_DGMLUTB                 0xE3   /* Digital Gamma LUT Blue */
#define ST7789V_GATECTRL                0xE4   /* Gate Control */
#define ST7789V_SPI2EN                  0xE7   /* SPI2 Enable */
#define ST7789V_PWCTRL2                 0xE8   /* Power Control 2 */
#define ST7789V_EQCTRL                  0xE9   /* Equalize time Control */
#define ST7789V_PROMCTRL                0xEC   /* Program Mode Control */
#define ST7789V_PROMEN                  0xFA   /* Program Mode Enable */
#define ST7789V_NVMSET                  0xFC   /* NVM Setting */
#define ST7789V_PROMACT                 0xFE   /* Program Action */

#define LCD_DMA_MAX_SIZE                (4095)

typedef struct {
  uint32_t clk_fre;
  uint8_t pin_clk;
  uint8_t pin_mosi;
  uint8_t pin_dc;
  uint8_t pin_cs;
  uint8_t pin_rst;
  uint8_t pin_bk;
  uint8_t horizontal;
  uint32_t max_buffer_size; // DMA used
  bool swap_data;
} lcd_config_t;

typedef struct {
  uint32_t buffer_size;
  uint32_t half_buffer_size;
  uint32_t node_cnt;
  uint32_t half_node_cnt;
  uint32_t dma_size;
  uint8_t pin_dc;
  uint8_t pin_cs;
  uint8_t pin_rst;
  uint8_t pin_bk;
  lldesc_t *dma;
  uint8_t *buffer;
  bool swap_data;
  QueueHandle_t event_queue;
} lcd_obj_t;

static lcd_obj_t *lcd_obj = NULL;

/* helper function to set the D/C line according the flag provided within user field */
/* Attention: function is called in context of SPI interrupt */
static void IRAM_ATTR lcd_dma_isr(void *arg)
{
  BaseType_t HPTaskAwoken = pdFALSE;
  typeof(GPSPI3.dma_int_st) int_st = GPSPI3.dma_int_st;
  GPSPI3.dma_int_clr.val = int_st.val;

  if (int_st.out_eof) {
    xQueueSendFromISR(lcd_obj->event_queue, (void *)&int_st.val, &HPTaskAwoken);
  }

  if (HPTaskAwoken == pdTRUE) {
    portYIELD_FROM_ISR();
  }
}

static void spi_write_data(uint8_t *data, size_t len)
{
  int event  = 0;
  int x = 0, cnt = 0, size = 0;
  int end_pos = 0;

  /*!< Generate a data DMA linked list */
  for (x = 0; x < lcd_obj->node_cnt; x++) {
    lcd_obj->dma[x].size = lcd_obj->dma_size;
    lcd_obj->dma[x].length = lcd_obj->dma_size;
    lcd_obj->dma[x].buf = (lcd_obj->buffer + lcd_obj->dma_size * x);
    lcd_obj->dma[x].eof = !((x + 1) % lcd_obj->half_node_cnt);
    lcd_obj->dma[x].empty = (uint32_t)&lcd_obj->dma[(x + 1) % lcd_obj->node_cnt];
  }

  lcd_obj->dma[lcd_obj->half_node_cnt - 1].empty = 0;
  lcd_obj->dma[lcd_obj->node_cnt - 1].empty = 0;
  cnt = len / lcd_obj->half_buffer_size;
  /*!< Start the signal */
  xQueueSend(lcd_obj->event_queue, &event, 0);

  /*!< Processing a complete piece of data, ping-pong operation */
  for (x = 0; x < cnt; x++) {
    uint8_t *out = (uint8_t *)lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt].buf;
    if (lcd_obj->swap_data) {
      for (size_t i = 0; i < lcd_obj->half_buffer_size; i += 2) {
        out[i]   = data[i+1];
        out[i+1] = data[i];
      }
    } else {
      memcpy(out, data, lcd_obj->half_buffer_size);
    }

    data += lcd_obj->half_buffer_size;
    xQueueReceive(lcd_obj->event_queue, (void *)&event, portMAX_DELAY);
    GPSPI3.mosi_dlen.usr_mosi_bit_len = lcd_obj->half_buffer_size * 8 - 1;
    GPSPI3.dma_out_link.addr = ((uint32_t)&lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt]) & 0xfffff;
    GPSPI3.dma_out_link.start = 1;
    ets_delay_us(1);
    GPSPI3.cmd.usr = 1;
  }

  cnt = len % lcd_obj->half_buffer_size;

  /*!< Processing remaining incomplete segment data */
  if (cnt) {
    uint8_t *out = (uint8_t *)lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt].buf;
    if (lcd_obj->swap_data && cnt > 1) {
      for (size_t i = 0; i < cnt; i += 2) {
        out[i]   = data[i+1];
        out[i+1] = data[i];
      }
    } else {
      memcpy(out, data, cnt);
    }

    /*!< Handle the case where the data length is an integer multiple of lcd_obj->dma_size */
    if (cnt % lcd_obj->dma_size) {
      end_pos = (x % 2) * lcd_obj->half_node_cnt + cnt / lcd_obj->dma_size;
      size = cnt % lcd_obj->dma_size;
    } else {
      end_pos = (x % 2) * lcd_obj->half_node_cnt + cnt / lcd_obj->dma_size - 1;
      size = lcd_obj->dma_size;
    }

    /*!< Handle the tail node to make it a DMA tail */
    lcd_obj->dma[end_pos].size = size;
    lcd_obj->dma[end_pos].length = size;
    lcd_obj->dma[end_pos].eof = 1;
    lcd_obj->dma[end_pos].empty = 0;
    xQueueReceive(lcd_obj->event_queue, (void *)&event, portMAX_DELAY);
    GPSPI3.mosi_dlen.usr_mosi_bit_len = cnt * 8 - 1;
    GPSPI3.dma_out_link.addr = ((uint32_t)&lcd_obj->dma[(x % 2) * lcd_obj->half_node_cnt]) & 0xfffff;
    GPSPI3.dma_out_link.start = 1;
    ets_delay_us(1);
    GPSPI3.cmd.usr = 1;
  }

  xQueueReceive(lcd_obj->event_queue, (void *)&event, portMAX_DELAY);
}


/* helper function to write a command register value in polling mode */
static void ST7789V_WriteCmnd( uint8_t aCmnd )
{
  gpio_set_level(lcd_obj->pin_dc, 0);
  spi_write_data(&aCmnd, 1);
}


/* helper function to read a data byte in polling mode */
static uint8_t ST7789V_ReadByte( void )
{
  return 0;
}


/* helper function to write a data byte in polling mode */
static void ST7789V_WriteByte( uint8_t aData )
{
  gpio_set_level(lcd_obj->pin_dc, 1);
  spi_write_data(&aData, 1);
}


/* helper function to write a data byte in polling mode */
static void ST7789V_WriteBytes(uint8_t *data, size_t len)
{
  gpio_set_level(lcd_obj->pin_dc, 1);
  spi_write_data(data, len);
}


/* helper function to initialize reset and backlight pins of LCD and to perform
   proper reset of display */
static void ST7789V_ResetDisplay()
{
  /* disable backlight */
  gpio_set_level( PIN_NUM_BCKL, 1 );

  /* apply hardware reset */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 5 / portTICK_PERIOD_MS );   /* reset signal asserted during 5ms  */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 20 / portTICK_PERIOD_MS );  /* reset signal asserted during 20ms */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_Init
*
* DESCRIPTION:
*   The function initializes the SPI hardware and all necessary GPIO in order
*   to reset and intialize the connected display hardware.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
static void lcd_config(lcd_config_t *config)
{
  REG_CLR_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_RST);
  REG_CLR_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_RST);
  REG_CLR_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_DMA_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_CLK_EN0_REG, DPORT_SPI3_DMA_CLK_EN);
  REG_SET_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_DMA_RST);
  REG_CLR_BIT(DPORT_PERIP_RST_EN0_REG, DPORT_SPI3_DMA_RST);

  int div = 2;

  if (config->clk_fre == 80000000) {
    GPSPI3.clock.clk_equ_sysclk = 1;
  } else {
    GPSPI3.clock.clk_equ_sysclk = 0;
    div = 80000000 / config->clk_fre;
  }

  GPSPI3.ctrl1.clk_mode = 0;
  GPSPI3.clock.clkdiv_pre = 1 - 1;
  GPSPI3.clock.clkcnt_n = div - 1;
  GPSPI3.clock.clkcnt_l = div - 1;
  GPSPI3.clock.clkcnt_h = ((div >> 1) - 1);

  GPSPI3.misc.ck_dis = 0;

  GPSPI3.user1.val = 0;
  GPSPI3.slave.val = 0;
  GPSPI3.misc.ck_idle_edge = 0;
  GPSPI3.user.ck_out_edge = 0;
  GPSPI3.ctrl.wr_bit_order = 0;
  GPSPI3.ctrl.rd_bit_order = 0;
  GPSPI3.user.val = 0;
  GPSPI3.user.cs_setup = 1;
  GPSPI3.user.cs_hold = 1;
  GPSPI3.user.usr_mosi = 1;
  GPSPI3.user.usr_mosi_highpart = 0;

  GPSPI3.dma_conf.val = 0;
  GPSPI3.dma_conf.out_rst = 1;
  GPSPI3.dma_conf.out_rst = 0;
  GPSPI3.dma_conf.ahbm_fifo_rst = 1;
  GPSPI3.dma_conf.ahbm_fifo_rst = 0;
  GPSPI3.dma_conf.ahbm_rst = 1;
  GPSPI3.dma_conf.ahbm_rst = 0;
  GPSPI3.dma_out_link.dma_tx_ena = 1;
  GPSPI3.dma_conf.out_eof_mode = 1;
  GPSPI3.cmd.usr = 0;

  GPSPI3.dma_int_clr.val = ~0;
  GPSPI3.dma_int_ena.val = 0;
  GPSPI3.dma_int_ena.out_eof = 1;

  intr_handle_t intr_handle = NULL;
  esp_intr_alloc(ETS_SPI3_DMA_INTR_SOURCE, 0, lcd_dma_isr, NULL, &intr_handle);
}

static void lcd_set_pin(lcd_config_t *config)
{
  PIN_FUNC_SELECT(GPIO_PIN_MUX_REG[config->pin_clk], PIN_FUNC_GPIO);
  gpio_set_direction(config->pin_clk, GPIO_MODE_OUTPUT);
  gpio_set_pull_mode(config->pin_clk, GPIO_FLOATING);
  gpio_matrix_out(config->pin_clk, SPI3_CLK_OUT_MUX_IDX, 0, 0);

  PIN_FUNC_SELECT(GPIO_PIN_MUX_REG[config->pin_mosi], PIN_FUNC_GPIO);
  gpio_set_direction(config->pin_mosi, GPIO_MODE_OUTPUT);
  gpio_set_pull_mode(config->pin_mosi, GPIO_FLOATING);
  gpio_matrix_out(config->pin_mosi, SPI3_D_OUT_IDX, 0, 0);

  /*!< Initialize non-SPI GPIOs */
  gpio_config_t io_conf;
  io_conf.intr_type = GPIO_PIN_INTR_DISABLE;
  io_conf.mode = GPIO_MODE_OUTPUT;
  io_conf.pin_bit_mask = (1ULL << config->pin_dc) | (1ULL << config->pin_cs);

  if (config->pin_rst <= 46) {
    io_conf.pin_bit_mask |= (1ULL << config->pin_rst);
  }

  if (config->pin_bk <= 46) {
    io_conf.pin_bit_mask |= (1ULL << config->pin_bk);
  }

  io_conf.pull_down_en = 0;
  io_conf.pull_up_en = 0;
  gpio_config(&io_conf);
}

void lcd_dma_config(lcd_config_t *config)
{
  int cnt = 0;

  if (config->max_buffer_size >= LCD_DMA_MAX_SIZE * 2) {
    lcd_obj->dma_size = LCD_DMA_MAX_SIZE;

    for (cnt = 0;; cnt++) { /*!< Find the buffer size that is divisible by dma_size */
      if ((config->max_buffer_size - cnt) % lcd_obj->dma_size == 0) {
        break;
      }
    }

    lcd_obj->buffer_size = config->max_buffer_size - cnt;
  } else {
    lcd_obj->dma_size = config->max_buffer_size / 2;
    lcd_obj->buffer_size = lcd_obj->dma_size * 2;
  }

  lcd_obj->half_buffer_size = lcd_obj->buffer_size / 2;

  lcd_obj->node_cnt = (lcd_obj->buffer_size) / lcd_obj->dma_size; /*!< Number of DMA nodes */
  lcd_obj->half_node_cnt = lcd_obj->node_cnt / 2;

  lcd_obj->dma  = (lldesc_t *)heap_caps_malloc(lcd_obj->node_cnt * sizeof(lldesc_t), MALLOC_CAP_DMA);
  lcd_obj->buffer = (uint8_t *)heap_caps_malloc(lcd_obj->buffer_size * sizeof(uint8_t), MALLOC_CAP_DMA);
}

int lcd_init(lcd_config_t *config)
{
  lcd_obj = (lcd_obj_t *)heap_caps_calloc(1, sizeof(lcd_obj_t), MALLOC_CAP_DMA);

  lcd_set_pin(config);
  lcd_config(config);
  lcd_dma_config(config);

  lcd_obj->event_queue = xQueueCreate(1, sizeof(int));

  lcd_obj->buffer_size = config->max_buffer_size;
  lcd_obj->swap_data = config->swap_data;

  lcd_obj->pin_dc = config->pin_dc;
  lcd_obj->pin_cs = config->pin_cs;
  lcd_obj->pin_rst = config->pin_rst;
  lcd_obj->pin_bk = config->pin_bk;
  gpio_set_level(lcd_obj->pin_cs, 1);

  /* perform reset of LCD */
  ST7789V_ResetDisplay();

  /* select chip */
  gpio_set_level(lcd_obj->pin_cs, 0);

  /* memory access control */
  ST7789V_WriteCmnd( ST7789V_MADCTL );
  ST7789V_WriteByte( 0x60 );

  /* interface pixel format, 16bits/pixel for RGB/MCU interface */
  ST7789V_WriteCmnd( ST7789V_COLMOD );
  ST7789V_WriteByte( 0x55 );

  /* porch setting */
  ST7789V_WriteCmnd( ST7789V_PORCTRL );
  ST7789V_WriteByte( 0x0C );
  ST7789V_WriteByte( 0x0C );
  ST7789V_WriteByte( 0x00 );
  ST7789V_WriteByte( 0x33 );
  ST7789V_WriteByte( 0x33 );

  /* gate control, Vgh=13.65V, Vgl=-10.43V */
  ST7789V_WriteCmnd( ST7789V_GCTRL );
  ST7789V_WriteByte( 0x45 );

  /* VCOM Setting, VCOM=1.175V */
  ST7789V_WriteCmnd( ST7789V_VCOMS );
  ST7789V_WriteByte( 0x2B );

  /* LCM Control, XOR: BGR, MX, MH */
  ST7789V_WriteCmnd( ST7789V_LCMCTRL );
  ST7789V_WriteByte( 0x2C );

  /* VDV and VRH Command Enable, enable=1 */
  ST7789V_WriteCmnd( ST7789V_VDVVRHEN );
  ST7789V_WriteByte( 0x01 );
  ST7789V_WriteByte( 0xFF );

  /* VRH Set, Vap=4.4+... */
  ST7789V_WriteCmnd( ST7789V_VRHS );
  ST7789V_WriteByte( 0x11 );

  /* VDV Set, VDV=0 */
  ST7789V_WriteCmnd( ST7789V_VDVS );
  ST7789V_WriteByte( 0x20 );

  /* frame rate control, 60Hz, inversion=0 */
  ST7789V_WriteCmnd( ST7789V_FRCTRL2 );
  ST7789V_WriteByte( 0x0F );

  /* power control 1, AVDD=6.8V, AVCL=-4.8V, VDDS=2.3V */
  ST7789V_WriteCmnd( ST7789V_PWCTRL1 );
  ST7789V_WriteByte( 0xA4 );
  ST7789V_WriteByte( 0xA1 );

  /* positive voltage gamma control */
  ST7789V_WriteCmnd( ST7789V_PVGAMCTRL );
  ST7789V_WriteByte( 0xD0 );
  ST7789V_WriteByte( 0x00 );
  ST7789V_WriteByte( 0x05 );
  ST7789V_WriteByte( 0x0E );
  ST7789V_WriteByte( 0x15 );
  ST7789V_WriteByte( 0x0D );
  ST7789V_WriteByte( 0x37 );
  ST7789V_WriteByte( 0x43 );
  ST7789V_WriteByte( 0x47 );
  ST7789V_WriteByte( 0x09 );
  ST7789V_WriteByte( 0x15 );
  ST7789V_WriteByte( 0x12 );
  ST7789V_WriteByte( 0x16 );
  ST7789V_WriteByte( 0x19 );

  /* negative voltage gamma control */
  ST7789V_WriteCmnd( ST7789V_NVGAMCTRL );
  ST7789V_WriteByte( 0xD0 );
  ST7789V_WriteByte( 0x00 );
  ST7789V_WriteByte( 0x05 );
  ST7789V_WriteByte( 0x0D );
  ST7789V_WriteByte( 0x0C );
  ST7789V_WriteByte( 0x06 );
  ST7789V_WriteByte( 0x2D );
  ST7789V_WriteByte( 0x44 );
  ST7789V_WriteByte( 0x40 );
  ST7789V_WriteByte( 0x0E );
  ST7789V_WriteByte( 0x1C );
  ST7789V_WriteByte( 0x18 );
  ST7789V_WriteByte( 0x16 );
  ST7789V_WriteByte( 0x19 );

  /* sleep out */
  ST7789V_WriteCmnd( ST7789V_SLPOUT );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* display on */
  ST7789V_WriteCmnd( ST7789V_DISPON );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* enable backlight */
  gpio_set_level( PIN_NUM_BCKL, 0 );

  return 0;
}

void ST7789V_Init( void )
{
  lcd_config_t lcd_config = {
    .clk_fre     = 40 * 1000 * 1000,
    .pin_clk     = PIN_NUM_CLK,
    .pin_mosi    = PIN_NUM_MOSI,
    .pin_dc      = PIN_NUM_DC,
    .pin_cs      = PIN_NUM_CS,
    .pin_rst     = PIN_NUM_RST,
    .pin_bk      = PIN_NUM_BCKL,
    .max_buffer_size = 2 * 1024,
    .swap_data     = 0,
  };

  lcd_init(&lcd_config);
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_ReadID
*
* DESCRIPTION:
*   The function assumes a connected and intialized ST7789V display and tries
*   to read its ID.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   If successful, the function returns the ID of the display - 0 otherwise.
*
*******************************************************************************/
uint16_t ST7789V_ReadID(void)
{
  uint16_t result;

  /* send command */
  ST7789V_WriteCmnd( ST7789V_RDDID );

  /* read dummy data */
  ST7789V_ReadByte();

  /* read register values */
  ST7789V_ReadByte();
  result = ST7789V_ReadByte() << 8;
  result |= ST7789V_ReadByte();

  return result;
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_WaitForCompletion
*
* DESCRIPTION:
*   The function ensures that the previous SPI data transfer is completed..
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ST7789V_WaitForCompletion( void )
{

}


/*******************************************************************************
* FUNCTION:
*   ST7789V_WriteDataWindow
*
* DESCRIPTION:
*   The function sets the destination position and size within the framebuffer
*   of the display according the given rectangle and starts the transfer of
*   the given data. The function does NOT wait for the completion of the transfer.
*
* ARGUMENTS:
*   aX       - Horizontal position for next write access.
*   aY       - Vertical position for next write access.
*   aW       - Width of the data window in pixel.
*   aH       - Height of the data window in pixel.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ST7789V_WriteDataWindow( uint16_t aX, uint16_t aY, uint16_t aW, uint16_t aH, uint8_t* aData )
{
  int       xEnd = aX + aW - 1;
  int       yEnd = aY + aH - 1;

  if ( xEnd >= DISPLAY_PIXEL_WIDTH )
    xEnd = DISPLAY_PIXEL_WIDTH - 1;
  if ( yEnd >= DISPLAY_PIXEL_HEIGHT )
    yEnd = DISPLAY_PIXEL_HEIGHT - 1;

  /* ensure that previous update is completed */
  ST7789V_WaitForCompletion();

  /* colomn address set */
  ST7789V_WriteCmnd(ST7789V_CASET);
  ST7789V_WriteByte(aX >> 8);
  ST7789V_WriteByte(aX & 0xFF);
  ST7789V_WriteByte(xEnd >> 8);
  ST7789V_WriteByte(xEnd & 0xFF);

  /* page address set */
  ST7789V_WriteCmnd(ST7789V_RASET);
  ST7789V_WriteByte(aY >> 8);
  ST7789V_WriteByte(aY & 0xFF);
  ST7789V_WriteByte(yEnd >> 8);
  ST7789V_WriteByte(yEnd & 0xFF);

  /* start writing to GRAM */
  ST7789V_WriteCmnd(ST7789V_RAMWR);
  ST7789V_WriteBytes((uint8_t *)aData, aW * aH * sizeof(uint16_t));
}


/* tzi, msy */